/*
    REMIND.C

Written by DavidW
Started on 6/12/91

*/

#include <p_std.h>
#include <p_gen.h>
#include <p_file.h>
#include <p_date.h>
#include <wlib.h>
#include <hwif.h>


LOCAL_D TEXT *cmds[]=
    {
    "iInsert",
    "cCopy",
    "uUpdate",
    "aAdd",
    "dDelete",
    "pPrint",
    "yPrint setup",
    "xExit",
    NULL
    };

#define EDIT_ITEMS      5
#define SPECIAL_ITEMS   3

#define ITEMS_BEFORE_EDIT       0
#define ITEMS_BEFORE_SPECIAL    EDIT_ITEMS

LOCAL_D H_MENU_DATA mdata[]=
    {
    "Edit",EDIT_ITEMS,
    "Special",SPECIAL_ITEMS,
    NULL
    };

GLDEF_D TEXT ** _cmds=(&cmds[0]);
GLDEF_D H_MENU_DATA * _mdata=(&mdata[0]);


LOCAL_D UINT MainWid;

typedef struct
    {
    ULONG sdate;
    UBYTE ref;
    UBYTE tlen;
    TEXT *pb;
    } ENTRY;

LOCAL_D ENTRY table[500];
LOCAL_D UWORD nent=0;
LOCAL_D INT cent=0;
LOCAL_D INT tent=0;

LOCAL_D ENTRY clipent;

LOCAL_D UWORD show_statwin=FALSE;
LOCAL_D UWORD ScreenWidth;

LOCAL_D UWORD lastbflags=W_BORD_CORNER_4;

LOCAL_D VOID *ttH;

LOCAL_D VOID *aH;
LOCAL_D WORD almstat;
LOCAL_D WORD almactive;

LOCAL_D UWORD PrintState;
LOCAL_D TEXT PrintBuf[48];
LOCAL_D UWORD PrintRec;


LOCAL_C INT cmp(INT r,VOID *pent)
    {
    return(table[r].sdate>(*(ULONG *)pent)? -1: 1);
    }

LOCAL_C INT LocateEntry(ULONG *psdate)
    {
    INT dir;
    INT ret;

    dir=p_bsrch(nent,cmp,&ret,psdate);
    if (dir>0)
        ret++;
    return(ret);
    }

LOCAL_C ENTRY *LocateRef(VOID)
    {
    ENTRY *pent;

    pent=(&table[0]);
    while (!pent->ref)
        pent++;
    return(pent);
    }

LOCAL_C VOID RequeueAlarm(VOID)
    {
    ULONG sdate;
    INT entno;
    ENTRY *pent;
    ULONG times[2];
    TEXT buf[66];

    if (almactive)
        {
        p_iow2(aH,P_FCANCEL);
        p_waitstat(&almstat);
        almactive=FALSE;
        LocateRef()->ref=FALSE;
        }
    if (!nent)
        return;
    sdate=p_date();
    entno=LocateEntry(&sdate);
    if (entno>=nent)
        return;
    pent=(&table[entno]);
    *p_bcpy(&buf[0],pent->pb,pent->tlen)=0;
    pent->ref=TRUE;
    times[0]=pent->sdate;
    times[1]=pent->sdate;
    p_ioc5(aH,P_FWRITE,&almstat,&times,&buf);
    almactive=TRUE;
    }

LOCAL_C VOID UpdateTimeFormat(VOID)
    {
    hTTSetFormat(ttH,H_TIME_FORMAT_NO_SECONDS|H_TIME_FORMAT_SUFFIX|
                            H_TIME_FORMAT_DAY_NAME|H_TIME_FORMAT_MONTH_NAME);
    }

LOCAL_C INT SdateToBuf(TEXT *pb,ULONG *psdate)
    {
    hTTSetTime(ttH,H_TIME_SET_SDATE,psdate);
    return(hTTSenseString(ttH,H_TIME_SENSE_BOTH,pb));
    }

LOCAL_C VOID SetGmode(INT gmode)
    {
    G_GC gc;

    gc.gmode=gmode;
    gSetGC(0,G_GC_MASK_GMODE,&gc);
    }

LOCAL_C VOID SetBold(INT flag)
    {
    G_GC gc;

    gc.font=WS_FONT_BASE+flag;
    gSetGC(0,G_GC_MASK_FONT,&gc);
    }

LOCAL_C VOID GetCursorBox(INT i,P_RECT *pbox)
    {
    pbox->tl.x=4;
    pbox->br.x=10;
    pbox->tl.y=4+i*18;
    pbox->br.y=pbox->tl.y+9;
    }

LOCAL_C VOID DrawCursor(INT i)
    {
    P_RECT box;

    GetCursorBox(i,&box);
    gPrintBoxText(&box,8,G_TEXT_ALIGN_LEFT,0,"\034",1);
    }

LOCAL_C VOID EraseCursor(INT i)
    {
    P_RECT box;

    GetCursorBox(i,&box);
    gClrRect(&box,G_TRMODE_CLR);
    }

LOCAL_C VOID LineToScreen(INT y,INT xoff,TEXT *pb,INT tlen)
    {
    P_RECT box;

    box.tl.x=4;
    box.tl.y=4+y*9;
    box.br.x=ScreenWidth-4;
    box.br.y=box.tl.y+9;
    gPrintBoxText(&box,8,G_TEXT_ALIGN_LEFT,xoff,pb,tlen);
    }

LOCAL_C VOID ClearLine(INT y)
    {
    LineToScreen(y,0,NULL,0);
    }

LOCAL_C VOID DrawLine(INT i)
    {
    INT len;
    INT entno;
    ENTRY *pent;
    TEXT buf[48];

    entno=i+tent;
    if (entno>=nent)
        {
        ClearLine(i*2);
        ClearLine(i*2+1);
        }
    else
        {
        pent=(&table[entno]);
        SetBold(TRUE);
        len=SdateToBuf(&buf[0],&pent->sdate);
        LineToScreen(i*2,10,&buf[0],len);
        SetBold(FALSE);
        if (entno==cent)
            DrawCursor(i);
        LineToScreen(i*2+1,22,pent->pb,pent->tlen);
        }
    }

LOCAL_C VOID Display(VOID)
    {
    INT i;

    if (!nent)
        {
        SetBold(TRUE);
        LineToScreen(0,10,"No reminders set",16);
        ClearLine(1);
        return;
        }
    for (i=0; i<4; i++)
        DrawLine(i);
    }

LOCAL_C INT ResizeWidth(UWORD width)
    {
    INT ret;
    W_WINDATA wd;

    wd.extent.tl.x=wd.extent.tl.y=0;
    wd.extent.width=width;
    wd.extent.height=80;
    wSetWindow(MainWid,W_WIN_EXTENT,&wd);
    ret=uErrorValue(wCheckPoint());
    if (!ret)
        ScreenWidth=width;
    return(ret);
    }

LOCAL_C INT ToggleStatusWindow(VOID)
    {
    if (show_statwin)
        {
        if (ResizeWidth(240))
            return(-1);
        wsDisable();
        }
    else
        {
        if (ResizeWidth(189))
            return(-1);
        wsEnable();
        }
    show_statwin^=TRUE;
    gBorder(lastbflags);
    Display();
    return(0);
    }

LOCAL_C VOID EnsureGoodTent(VOID)
    {
    if (nent<5)
        tent=0;
    else if (tent<cent-3)
        tent=cent-3;
    else if (tent>cent)
        tent=cent;
    }

LOCAL_C VOID AppendEntry(ENTRY *pent)
    {
    INT tomove;

    if (!nent)
        cent=0;
    else
        {
        cent=LocateEntry(&pent->sdate);
        tomove=nent-cent;
        p_bcpy(&table[cent+1],&table[cent],tomove*sizeof(ENTRY));
        }
    nent++;
    table[cent]=(*pent);
    EnsureGoodTent();
    Display();
    RequeueAlarm();
    }

LOCAL_C VOID DoDeleteEntry(VOID)
    {
    INT tomove;

    p_free(table[cent].pb);
    nent--;
    tomove=nent-cent;
    p_bcpy(&table[cent],&table[cent+1],tomove*sizeof(ENTRY));
    }

LOCAL_C VOID CleanEntry(ENTRY *pent)
    {
    p_bfil(pent,sizeof(ENTRY),0);
    pent->sdate=p_date()+3600;
    }

LOCAL_C INT EditEntry(ENTRY *pent)
    {
    P_DAYSEC ds;
    H_DI_DATE date;
    H_DI_TIME time;
    TEXT buf[66];
    H_DI_SEDIT sed;

    if (uOpenDialog("Reminder"))
        return(-1);
    p_bcpy(&buf[1],pent->pb,buf[0]=pent->tlen);
    sed.str=(&buf[0]);
    sed.len=64;
    sed.width=25;
    if (uAddDialogItem(H_DIALOG_SEDIT,"Message",&sed))
        return(-1);
    p_sttods(&pent->sdate,&ds);
    time.value=(&ds.sec);
    time.low=0;
    time.high=P_NSECDAY-1;
    time.type=0;
    if (uAddDialogItem(H_DIALOG_TIME,"Time",&time))
        return(-1);
    date.value=(&ds.day);
    date.low=H_FIRST_SYS_DAY;
    date.high=H_LAST_SYS_DAY;
    if (uAddDialogItem(H_DIALOG_DATE,"Date",&date))
        return(-1);
    if (uRunDialog()<=0)
        return(-1);
    p_dstost(&ds,&pent->sdate);
    pent->pb=p_realloc(pent->pb,buf[0]);
    if (!pent->pb)
        {
        uErrorValue(E_GEN_NOMEMORY);
        return(-1);
        }
    pent->tlen=buf[0];
    p_bcpy(pent->pb,&buf[1],buf[0]);
    return(0);
    }

LOCAL_C VOID AddEntry(VOID)
    {
    ENTRY ent;

    CleanEntry(&ent);
    if (EditEntry(&ent))
        return;
    AppendEntry(&ent);
    }

LOCAL_C VOID UpdateEntry(VOID)
    {
    ENTRY *pent;
    ENTRY ent;

    if (!nent)
        {
        wInfoMsg("No entry to update");
        return;
        }
    pent=(&table[cent]);
    if (EditEntry(pent))
        return;
    ent=(*pent);
    pent->pb=0;
    DoDeleteEntry();
    AppendEntry(&ent);
    }

LOCAL_C VOID DeleteEntry(VOID)
    {
    if (!nent)
        {
        wInfoMsg("No entry to delete");
        return;
        }
    if (uOpenDialog("Delete this reminder"))
        return;
    if (uAddButtonList("No",-'n',"Yes",'y',NULL))
        return;
    if (uRunDialog()!='y')
        return;
    DoDeleteEntry();
    if (cent==nent && nent)
        cent=nent-1;
    EnsureGoodTent();
    Display();
    RequeueAlarm();
    }

LOCAL_C INT TakeCopy(ENTRY *targ,ENTRY *src)
    {
    targ->pb=p_alloc(src->tlen);
    if (!targ->pb)
        {
        uErrorValue(E_GEN_NOMEMORY);
        return(-1);
        }
    p_bcpy(targ->pb,src->pb,src->tlen);
    p_bcpy(targ,src,sizeof(ENTRY)-sizeof(TEXT *));
    return(0);
    }

LOCAL_C VOID CopyEntry(VOID)
    {
    ENTRY new;

    if (!nent)
        {
        wInfoMsg("No entry to copy");
        return;
        }
    if (TakeCopy(&new,&table[cent]))
        return;
    p_free(clipent.pb);
    clipent=new;
    wInfoMsg("Reminder copied");
    }

LOCAL_C VOID InsertEntry(VOID)
    {
    ENTRY new;

    if (!clipent.pb)
        {
        wInfoMsg("No entry to insert");
        return;
        }
    if (TakeCopy(&new,&clipent))
        return;
    AppendEntry(&new);
    }

LOCAL_C VOID ExitIfConfirmed(VOID)
    {
    H_DI_TEXT txt;
    TEXT buf[40];

    if (nent)
        {
        if (uOpenDialog("Warning"))
            return;
        uZTStoBCS(&buf[0],"Exiting will delete all reminders");
        txt.str=(&buf[0]);
        txt.type=H_DTEXT_ALIGN_CENTRE;
        if (uAddDialogItem(H_DIALOG_TEXT,0,&txt))
            return;
        if (uAddButtonList("Cancel",W_KEY_ESCAPE,"Exit",W_KEY_RETURN,NULL))
            return;
        if (uRunDialog()!=W_KEY_RETURN)
            return;
        }
    p_exit(0);
    }

LOCAL_C INT PrintLine(H_PRINT *pr)
    {
    ENTRY *pent;

    pent=(&table[PrintRec]);
    switch (PrintState++)
        {
    case 0:
        pr->blen=SdateToBuf(&PrintBuf[0],&pent->sdate);
        pr->buf=(&PrintBuf[0]);
        pr->style|=2;
        break;
    case 1:
        pr->buf=pent->pb;
        pr->blen=pent->tlen;
        break;
    case 2:
        pr->blen=0;
        PrintState=0;
        PrintRec++;
        return(PrintRec!=nent);
        }
    pr->flags|=0x8000;
    return(TRUE);
    }

LOCAL_C VOID PrintAll(VOID)
    {
    if (!nent)
        {
        wInfoMsg("Nothing to print");
        return;
        }
    PrintRec=0;
    PrintState=0;
    hPrint(PrintLine);
    }

LOCAL_C VOID ManageCommand(INT index)
    {
    switch (index)
        {
    case ITEMS_BEFORE_EDIT:
        InsertEntry();
        break;
    case ITEMS_BEFORE_EDIT+1:
        CopyEntry();
        break;
    case ITEMS_BEFORE_EDIT+2:
        UpdateEntry();
        break;
    case ITEMS_BEFORE_EDIT+3:
        AddEntry();
        break;
    case ITEMS_BEFORE_EDIT+4:
        DeleteEntry();
        break;
    case ITEMS_BEFORE_SPECIAL:
        PrintAll();
        break;
    case ITEMS_BEFORE_SPECIAL+1:
        hPrintSetupDialog();
        break;
    case ITEMS_BEFORE_SPECIAL+2:
        ExitIfConfirmed();
        break;
        }
    }

LOCAL_C VOID TryExecuteCommand(INT keycode)
    {
    keycode=uLocateCommand(keycode);
    if (keycode>=0)
        ManageCommand(keycode);
    }

LOCAL_C VOID DoScroll(INT dely)
    {
    P_POINT offset;
    P_RECT src;

    offset.x=0;
    offset.y=dely*18;
    src.tl.x=4;
    src.br.x=ScreenWidth-4;
    if (dely>0)
        {
        src.tl.y=4;
        src.br.y=4+(4-dely)*18;
        }
    else
        {
        src.tl.y=4-dely*18;
        src.br.y=4+4*18;
        }
    wScrollRect(MainWid,&src,&offset);
    }

LOCAL_C VOID MakeHighlight(INT newcent)
    {
    INT toscroll;

    if (!nent)
        return;
    if (newcent==cent)
        return;
    EraseCursor(cent-tent);
    toscroll=tent-newcent;
    if (toscroll>0)
        {
        tent-=toscroll;
        if (toscroll<4)
            DoScroll(toscroll);
        else
            toscroll=4;
        while (toscroll)
            DrawLine(--toscroll);
        }
    else
        {
        toscroll=-3-toscroll;
        if (toscroll>0)
            {
            tent+=toscroll;      
            if (toscroll<4)
                DoScroll(-toscroll);
            else
                toscroll=4;
            while (toscroll)
                DrawLine(4-toscroll--);
            }
        }
    cent=newcent;
    DrawCursor(cent-tent);
    }

LOCAL_C VOID ScrollPageBy(INT delta)
    {
    if (!nent)
        return;
    if (delta<(-3))
        delta=(-3);
    else if (delta>3)
        delta=3;
    tent+=delta;
    cent+=delta;
    Display();
    }

LOCAL_C VOID HandleAlarmEvent(VOID)
    {
    ENTRY *pent;
    INT expired;

    pent=LocateRef();
    expired=(pent->sdate<=p_date());
    almactive=FALSE;
    pent->ref=FALSE;
    RequeueAlarm();
    if (expired)
        {
        uForceToFront();
        MakeHighlight(pent-(&table[0]));
        UpdateEntry();
        }
    }

LOCAL_C VOID CheckBorderArrows(VOID)
    {
    INT bflags;

    if (!nent)
        return;
    if (tent)
        bflags=W_BORD_CORNER_4|W_BORD_TOP_ON;
    else
        bflags=W_BORD_CORNER_4|W_BORD_TOP_OFF;
    if (tent+4<nent)
        bflags|=W_BORD_BOT_ON;
    else
        bflags|=W_BORD_BOT_OFF;
    if (bflags!=lastbflags)
        gBorder(lastbflags=bflags);
    }

LOCAL_C VOID MainLoop(VOID)
    {
    INT ret;
    WMSG_KEY key;
    WORD keystat;
    WORD keyactive;

    keyactive=FALSE;
    FOREVER
        {
        CheckBorderArrows();
        if (keyactive)
            wFlush();
        else
            {
            uGetKeyA(&keystat,&key);
            keyactive=TRUE;
            }
        p_iowait();
        if (keystat==E_FILE_PENDING)
            {
            HandleAlarmEvent();
            continue;
            }
        keyactive=FALSE;
        if (key.keycode&W_EVENT_KEY)
            {
            if (key.keycode==CONS_EVENT_FOREGROUND)
                {
                UpdateTimeFormat();
                Display();
                }
            }
        else if (key.keycode&W_SPECIAL_KEY)
            TryExecuteCommand(key.keycode&(~W_SPECIAL_KEY));
        else
            {
            switch (key.keycode)
                {
            case W_KEY_MENU:
                if (key.modifiers&W_CTRL_MODIFIER)
                    {
                    SetGmode(G_TRMODE_CLR);
                    gBorder(lastbflags&~(W_BORD_TOP_OFF|W_BORD_BOT_OFF));
                    SetGmode(G_TRMODE_SET);
                    ToggleStatusWindow();
                    }
                else
                    {
                    ret=uPresentMenus();
                    if (ret>0)
                        TryExecuteCommand(ret);
                    }
                break;
            case W_KEY_DELETE_LEFT:
                DeleteEntry();
                break;
            case W_KEY_RETURN:
                if (!nent || key.modifiers&W_SHIFT_MODIFIER) 
                    AddEntry();
                else
                    UpdateEntry();
                break;
            case W_KEY_UP:
                if (cent)
                    MakeHighlight(cent-1);
                break;
            case W_KEY_DOWN:
                if (cent<nent-1)
                    MakeHighlight(cent+1);
                break;
            case W_KEY_PAGE_UP:
                if (key.modifiers&W_CTRL_MODIFIER)
                    MakeHighlight(0);
                else if (!tent)
                    MakeHighlight(0);
                else
                    ScrollPageBy(-tent);
                break;
            case W_KEY_PAGE_DOWN:
                if (key.modifiers&W_CTRL_MODIFIER)
                    MakeHighlight(nent-1);
                else if (tent+4>=nent)
                    MakeHighlight(nent-1);
                else
                    ScrollPageBy(nent-tent-4);
                break;
                }
            }
        }
    }

LOCAL_C VOID CreateGC(VOID)
    {
    INT hgc;

    hgc=gCreateGC0(MainWid);
    if (hgc<0)
        p_exit(hgc);
    }

LOCAL_C VOID SpecificInit(VOID)
    {
    uEscape(FALSE);
    MainWid=uFindMainWid();
    CreateGC();
    p_open(&aH,"ALM:",-1);
    ttH=hTTOpen();
    hTTSetAbbreviations(ttH,3,3);
    UpdateTimeFormat();
    hCrackCommandLine();
    if (ToggleStatusWindow())
        p_exit(0);
    }

GLDEF_C VOID main(VOID)
    {
    uCommonInit();
    SpecificInit();
    MainLoop();
    }
